#region License
/*
MIT License
Copyright © 2006 The Mono.Xna Team

Adapted for Vortex2D.NET
 * Alex Khomich
 
All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/
#endregion License

using System;
using System.Collections.Generic;
using System.Text;

namespace Vortex.Drawing {

	/// <summary>
	/// Container of mathematic helper functions and constants
	/// </summary>
	public static class MathHelper {
		///<summary>E itself</summary>
		public const float E = (float)Math.E;
		///<summary>Logarithm of E with 10 basis</summary>
		public const float Log10E = 0.4342945f;
		///<summary>Logarithm of E with 2 basis</summary>
		public const float Log2E = 1.442695f;
		///<summary>PI itself</summary>
		public const float Pi = (float)Math.PI;
		///<summary>A half of PI</summary>
		public const float PiOver2 = (float)(Math.PI / 2.0);
		///<summary>1/4th of PI</summary>
		public const float PiOver4 = (float)(Math.PI / 4.0);
		///<summary>2 * PI</summary>
		public const float TwoPi = (float)(Math.PI * 2.0);

		public static float Barycentric(float value1, float value2, float value3, float amount1, float amount2) {
			return value1 + (value2 - value1) * amount1 + (value3 - value1) * amount2;
		}

		public static float CatmullRom(float value1, float value2, float value3, float value4, float amount) {
			// Using formula from http://www.mvps.org/directx/articles/catmull/
			// Internally using doubles not to lose precission
			double amountSquared = amount * amount;
			double amountCubed = amountSquared * amount;
			return (float)(0.5 * (2.0 * value2 +
				(value3 - value1) * amount +
				(2.0 * value1 - 5.0 * value2 + 4.0 * value3 - value4) * amountSquared +
				(3.0 * value2 - value1 - 3.0 * value3 + value4) * amountCubed));
		}

		///<summary>Clamp float value with specified min/max</summary>
		public static float Clamp(float value, float min, float max) {
			// First we check to see if we're greater than the max
			value = (value > max) ? max : value;

			// Then we check to see if we're less than the min.
			value = (value < min) ? min : value;

			// There's no check to see if min > max.
			return value;
		}

		///<summary>Calculates absolute distance between two values in 1D</summary>
		public static float Distance(float value1, float value2) {
			return Math.Abs(value1 - value2);
		}

		public static float Hermite(float value1, float tangent1, float value2, float tangent2, float amount) {
			// All transformed to double not to lose precission
			// Otherwise, for high numbers of param:amount the result is NaN instead of Infinity
			double v1 = value1, v2 = value2, t1 = tangent1, t2 = tangent2, s = amount, result;
			double sCubed = s * s * s;
			double sSquared = s * s;

			if (amount == 0f)
				result = value1;
			else if (amount == 1f)
				result = value2;
			else
				result = (2 * v1 - 2 * v2 + t2 + t1) * sCubed +
					(3 * v2 - 3 * v1 - 2 * t1 - t2) * sSquared +
					t1 * s +
					v1;
			return (float)result;
		}

		///<summary>Calculates linear interpolated value between two values in 1D</summary>
		public static float Lerp(float value1, float value2, float amount) {
			return value1 + (value2 - value1) * amount;
		}

		///<summary>Return greater value</summary>
		public static float Max(float value1, float value2) {
			return Math.Max(value1, value2);
		}

		///<summary>Return less value</summary>
		public static float Min(float value1, float value2) {
			return Math.Min(value1, value2);
		}

		public static float SmoothStep(float value1, float value2, float amount) {
			// It is expected that 0 < amount < 1
			// If amount < 0, return value1
			// If amount > 1, return value2
			float result = MathHelper.Clamp(amount, 0f, 1f);
			result = MathHelper.Hermite(value1, 0f, value2, 0f, result);
			return result;
		}

		/// <summary>
		/// Converts radians into degrees.
		/// </summary>
		/// <param name="radians">Angle in radians.</param>
		/// <returns>Angle in degrees.</returns>
		public static float ToDegrees(float radians) {
			// This method uses double precision internally,
			// though it returns single float
			// Factor = 180 / pi
			return (float)(radians * 57.295779513082320876798154814105);
		}

		/// <summary>
		/// Converts degrees into radians.
		/// </summary>
		/// <param name="degrees">Angle in degrees.</param>
		/// <returns>Angle in radians</returns>
		public static float ToRadians(float degrees) {
			// This method uses double precision internally,
			// though it returns single float
			// Factor = pi / 180
			return (float)(degrees * 0.017453292519943295769236907684886);
		}


		/// <summary>
		/// Wraps the angle in range -PI .. +PI.
		/// </summary>
		/// <param name="angle">The angle in radians</param>
		/// <returns>Wrapped angle in radians in range -PI .. +PI</returns>
		public static float WrapAngle(float angle) {
			angle = (float)Math.IEEERemainder((double)angle, 6.2831854820251465); //2xPi precision is double
			if (angle <= -3.141593f) {
				angle += 6.283185f;
				return angle;
			}
			if (angle > 3.141593f) {
				angle -= 6.283185f;
			}
			return angle;
		}

		/// <summary>
		/// Calculates square root the specified value.
		/// </summary>
		/// <param name="value">The float value</param>
		/// <returns>Squared value</returns>
		public static float Sqrt(float value) {
			return (float)Math.Sqrt(value);
		}

		public static float Floor(float value) {
			return (float)Math.Floor(value);
		}

		public static float Ceiling(float value) {
			return (float)Math.Ceiling(value);
		}

		public static float Round(float value) {
			return (float)Math.Round(value);
		}

		public static float Round(float value, int decimals) {
			return (float)Math.Round(value, decimals);
		}
	}
}
