﻿using System;
using System.Collections.Generic;
using System.Text;
using Vortex.Wrapper.Direct3D;
using Vortex.Wrapper;
using Vortex.Debugging;

namespace Vortex.Drawing {
	/// <summary>
	/// Pixel shader object. Stateful, contains shader params
	/// </summary>
	public class PixelShader : Shader {
		private Direct3DPixelShader9 _shader;

		#region Constructors
		/// <summary>
		/// Initializes a new instance of the <see cref="PixelShader"/> class.
		/// </summary>
		/// <param name="fileName">Name of the file with shader code.</param>
		/// <param name="profile">The pixel shader compilation profile.</param>
        /// <param name="entryPoint">The name of entry point function for this shader.</param>
        public PixelShader(string fileName, PixelShaderProfile profile, string entryPoint) {
            using (D3DXBuffer shaderBytes = CompileFromFile(fileName, MakeProfileString(profile), entryPoint)) {
                unsafe
                {
                    HResult result = Device.ApiRef.CreatePixelShader(shaderBytes.GetBufferPointer(), out _shader);
                    Watcher.AssertCritical(result, "Create pixel shader from file: '{0}', profile: {1}, entryPoint: {2}", fileName, profile, entryPoint);
                }
            }
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="PixelShader"/> class. Entry point - default.
		/// </summary>
		/// <param name="fileName">Name of the file with shader code.</param>
		/// <param name="profile">The ps profile.</param>
		public PixelShader(string fileName, PixelShaderProfile profile) :
			this(fileName, profile, DefaultEntryPoint){
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="PixelShader"/> class. Profile - default.
		/// </summary>
		/// <param name="fileName">Name of the file with shader code.</param>
		/// <param name="entryPoint">The entry point fn name.</param>
		public PixelShader(string fileName, string entryPoint) :
			this(fileName, DefaultProfile, entryPoint) {
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="PixelShader"/> class. Profile and entry point are default.
		/// </summary>
		/// <param name="fileName">Name of the file with shader code.</param>
		public PixelShader(string fileName) :
			this(fileName, DefaultProfile, DefaultEntryPoint) {
		}

		#endregion

		/// <summary>
		/// Gets the shader wrapper object.
		/// </summary>
		/// <value>The wrapper object.</value>
		internal Direct3DPixelShader9 Handle {
			get { return _shader; }
		}

		/// <summary>
		/// Converts the pixel shader profile into  string.
		/// </summary>
		/// <param name="profile">The ps profile.</param>
		/// <returns>String representation of profile</returns>
		private static string MakeProfileString(PixelShaderProfile profile) {
			int value = (int)profile;
			return string.Format("ps_{0}_{1}", (value >> 16), (value & 0xFFFF));
		}

		/// <summary>
		/// Frees the internal data of shader.
		/// </summary>
		protected override void FreeData() {
			_shader.Dispose();
			base.FreeData();
		}

		/// <summary>
		/// Gets the default entry point for pixel shader.
		/// </summary>
		/// <value>The default entry point name.</value>
		public static string DefaultEntryPoint {
			get { return "ps_main"; }
		}

		/// <summary>
		/// Gets the default profile for pixel shader.
		/// </summary>
		/// <value>The ps profile.</value>
		public static PixelShaderProfile DefaultProfile {
			get { return PixelShaderProfile.Ps_2_0; }
		}

		/// <summary>Default pixel shader instance</summary>
		public static readonly PixelShader Default = null;
	}
}
